﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using ManagedDirectX;
using ManagedDirectX.Direct3D11;
using ManagedDirectX.Direct3D11.Resource11;
using ManagedDirectX.DXGI;

namespace D3D11Helper
{
    /// <summary>
    /// Classe Gestione Device DirectX10
    /// </summary>
    public class DeviceManager
    {
        /// <summary>
        /// Puntatore al device
        /// </summary>
        public Device DevicePointer { get; private set; }
        /// <summary>
        /// puntatore allo SwapChain
        /// </summary>
        public GISwapChain SwapChainPointer { get; private set; }

        /// <summary>
        /// Contesto
        /// </summary>
        public DeviceContext DeviceContextPointer { get; private set; }

        /// <summary>
        /// Puntatore al render target principale
        /// </summary>
        public RenderTargetView RenderTargetPointer { get; private set; }
        /// <summary>
        /// Puntatore al depth stencil
        /// </summary>
        public DepthStencilView DepthStencilPointer { get; private set; }

        /// <summary>
        /// Blend state
        /// </summary>
        public BlendState BlendState { get; private set; }

        /// <summary>
        /// Raster state
        /// </summary>
        public RasterizerState RasterState { get; private set; }

        /// <summary>
        /// Depth state
        /// </summary>
        public DepthStencilState DepthState { get; private set; }

        /// <summary>
        /// Sampler state
        /// </summary>
        public SamplerState SamplerState { get; private set; }


        /// <summary>
        /// Costruttore
        /// </summary>
        /// <param name="handle">handle del controllo</param>
        /// <param name="width">larghezza</param>
        /// <param name="heigth">altezza</param>
        /// <param name="windowed">true per utilizzare la modalità in finestra</param>
        /// <param name="debugMode">modalità debug</param>
        public DeviceManager(IntPtr handle, int width, int heigth, bool windowed, bool debugMode)
        {
            if (windowed)
                CreateWindowedDevice(handle, width, heigth, debugMode);
            CreateTargetView(width, heigth);
            SetTargets(width, heigth);


            DepthStencilDescription descD = new DepthStencilDescription();
            descD.DepthEnable = true;
            descD.DepthWriteMask = DepthWriteMask.All;
            descD.DepthFunction = ComparisonFunction.Less_Equal;

            RasterizerDescription rastD = new RasterizerDescription();
            rastD.CullMode = CullMode.Back;
            rastD.FillMode = FillMode.Solid;
            rastD.ScissorEnable = false;

            SamplerDescription sampD = new SamplerDescription();
            sampD.AddressU = TextureAddressMode.Wrap;
            sampD.AddressV = TextureAddressMode.Wrap;
            sampD.AddressW = TextureAddressMode.Wrap;
            sampD.Filter = Filter.MinMagMipLinear;

            BlendDescription blenD = new BlendDescription();
            blenD.IndependentBlendEnable = false;
            blenD.RenderTarget = new RenderTargetBlendDescription[8];
            blenD.RenderTarget[0].SrcBlend = Blend.One;
            blenD.RenderTarget[0].DestBlend = Blend.Zero;
            blenD.RenderTarget[0].BlendOp = BlendOperation.Add;
            blenD.RenderTarget[0].SrcBlendAlpha = Blend.One;
            blenD.RenderTarget[0].DestBlendAlpha = Blend.Zero;
            blenD.RenderTarget[0].BlendOpAlpha = BlendOperation.Add;
            blenD.RenderTarget[0].RenderTargetWriteMask = 15;


            RasterState = DevicePointer.CreateRasterizerState(rastD);
            DepthState = DevicePointer.CreateDepthStencilState(descD);
            SamplerState = DevicePointer.CreateSamplerState(sampD);
            BlendState = DevicePointer.CreateBlendState(blenD);

            DeviceContextPointer.OMSetBlendState(BlendState, new float[] { 0, 0, 0, 0 });
            DeviceContextPointer.RSSetState(RasterState);
            DeviceContextPointer.OMSetDepthStencilState(DepthState);
            DeviceContextPointer.PSSetSamplers(new SamplerState[] { SamplerState });
        }

        /// <summary>
        /// Costruttore
        /// </summary>
        /// <param name="handle">handle del controllo</param>
        /// <param name="width">larghezza</param>
        /// <param name="heigth">altezza</param>
        public DeviceManager(IntPtr handle, int width, int heigth)
            : this(handle, width, heigth, true, false)
        {

        }

        private void CreateWindowedDevice(IntPtr handle, int width, int heigth, bool debug)
        {
            SwapChainDescription sd = new SwapChainDescription();
            sd.BufferCount = 1;
            sd.BufferDescription = new ModeDescription(width, heigth, new Rational(60, 1), Format.R8G8B8A8_UNORM);
            sd.BufferUsage = ManagedDirectX.DXGI.Usage.RenderTargetOutput;
            sd.SampleDescription = new SampleDescription(1, 0);
            sd.Windowed = true;
            sd.OutputWindow = handle;

            GISwapChain swap;
            DeviceContext context;
            
            DevicePointer = Device.CreateDeviceAndSwapChain((debug) ? DriverType.Reference : DriverType.Hardware, sd, out swap, out context);
            SwapChainPointer = swap;
            DeviceContextPointer = context;
        }


        /// <summary>
        /// Pulisce il target principale
        /// </summary>
        /// <param name="red">rosso</param>
        /// <param name="green">verde</param>
        /// <param name="blue">blu</param>
        /// <param name="alpha">trasparenza</param>
        public void Clear(float red, float green, float blue, float alpha)
        {
            DeviceContextPointer.ClearRenderTargetView(RenderTargetPointer, red, green, blue, alpha);
        }

        /// <summary>
        /// Pulisce il depth stencil
        /// </summary>
        /// <param name="clearDepth">true per pulire il depth</param>
        /// <param name="clearStencil">true per pulire lo stencil</param>
        /// <param name="depth">valore di depth</param>
        /// <param name="stencil">valore di stencil</param>
        public void ClearDepth(bool clearDepth, bool clearStencil, float depth, byte stencil)
        {
            ClearFlag flag = 0;
            if (clearDepth)
                flag = flag | ClearFlag.Depth;
            if (clearStencil)
                flag = flag | ClearFlag.Stencil;

            DeviceContextPointer.ClearDepthStencilView(DepthStencilPointer, flag, depth, stencil);
        }



        /// <summary>
        /// Invia la scena allo schermo
        /// </summary>
        public void Present()
        {
            SwapChainPointer.Present();
        }

        /// <summary>
        /// Corregge la dimensione del device
        /// </summary>
        /// <param name="width">Larghezza</param>
        /// <param name="height">Altezza</param>
        public void Resize(int width, int height)
        {
            DeviceContextPointer.ClearState();
            this.DepthStencilPointer.Dispose();
            this.RenderTargetPointer.Dispose();

            while (DepthStencilPointer.IsValid)
            {
            }

            while (RenderTargetPointer.IsValid)
            {
            }

            SwapChainPointer.ResizeBuffers(1, width, height, Format.R8G8B8A8_UNORM, SwapChainFlag.AllowModeSwitch);
            CreateTargetView(width, height);
            SetTargets(width, height);

            SetDefaultState();
        }

        /// <summary>
        /// Imposta lo stato di default nel device
        /// </summary>
        public void SetDefaultState()
        {
            DeviceContextPointer.OMSetBlendState(BlendState, new float[] { 0, 0, 0, 0 });
            DeviceContextPointer.OMSetDepthStencilState(DepthState);
            DeviceContextPointer.PSSetSamplers(new SamplerState[] { SamplerState });
            DeviceContextPointer.RSSetState(RasterState);

            RasterizerDescription desc = RasterState.GetDescription();
            DeviceContextPointer.RSSetScissorRects(new Rectangle[] { new Rectangle(0, 0, 640, 480) });
        }



        private void CreateTargetView(int width, int height)
        {
            Texture2D pBackBuffer = Texture2D.GetBuffer(SwapChainPointer);
            RenderTargetPointer = DevicePointer.CreateRenderTargetView(pBackBuffer);
            pBackBuffer.Dispose();
            DepthStencilPointer = CreateDepthStencil(width, height);
        }

        private DepthStencilView CreateDepthStencil(int width, int heigth)
        {

            Texture2DDescription descDepth = new Texture2DDescription();

            descDepth.Width = width;
            descDepth.Height = heigth;
            descDepth.MipLevels = 1;
            descDepth.ArraySize = 1;

            descDepth.Format = Format.D32_FLOAT;

            descDepth.SampleDescription = new SampleDescription(1, 0);

            descDepth.Usage = ManagedDirectX.Direct3D11.Resource11.Usage.Default;
            descDepth.BindFlags = BindingFlag.DepthStencil;
            descDepth.CPUAccessFlags = 0;
            descDepth.MiscFlags = 0;

            Texture2D pDepthBuffer = DevicePointer.CreateTexture2D(descDepth);

            DepthStencilViewDescription descDSV = new DepthStencilViewDescription();
            descDSV.Format = descDepth.Format;
            descDSV.ViewDimension = DepthStencilViewDimension.Texture2D;
            descDSV.Texture2D = new Texture2DDepthStencilView();

            DepthStencilView view = DevicePointer.CreateDepthStencilView(pDepthBuffer, descDSV);
            pDepthBuffer.Dispose();
            return view;
        }


        private void SetTargets(int width, int heigth)
        {

            DeviceContextPointer.OMSetRenderTargets(RenderTargetPointer, DepthStencilPointer);
            DeviceContextPointer.RSSetViewports(width, heigth);
        }


    }

}

