﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using ManagedDirectX;
using ManagedDirectX.DXGI;
using ManagedDirectX.Direct3D11;
using ManagedDirectX.Direct3D11.Resource11;
using Buffer11 = ManagedDirectX.Direct3D11.Resource11.Buffer;

namespace D3D11Helper
{

    /// <summary>
    /// 3D Model
    /// </summary>
    public class Object3D
    {
        /// <summary>
        /// Vertex buffer
        /// </summary>
        public Buffer11 VertexBuffer { get; private set; }

        /// <summary>
        /// Index buffer
        /// </summary>
        public Buffer11 IndexBuffer { get; private set; }

        /// <summary>
        /// Layout
        /// </summary>
        public InputLayout Layout { get; private set; }

        /// <summary>
        /// Vertex Stride
        /// </summary>
        public int VertexStride { get; private set; }

        /// <summary>
        /// Face count
        /// </summary>
        public int FaceCount { get; private set; }

        /// <summary>
        /// Material
        /// </summary>
        public List<Material3D> Materials { get; private set; }

        /// <summary>
        /// Static creator
        /// </summary>
        /// <typeparam name="VertexType">Vertex Type</typeparam>
        /// <param name="vertices">Vertices</param>
        /// <param name="indices">Indices</param>
        /// <param name="desc">Description</param>
        /// <param name="effect">Effect</param>
        /// <param name="device">Device</param>
        /// <returns>Object3D</returns>
        public static Object3D Create<VertexType>(VertexType[] vertices, short[] indices, InputElementDescription[] desc, EffectShader effect, DeviceManager device)
            where VertexType : struct
        {
            Object3D obj = new Object3D();
            BufferDescription bd = new BufferDescription();

            bd.Usage = ManagedDirectX.Direct3D11.Resource11.Usage.Default;
            obj.VertexStride = System.Runtime.InteropServices.Marshal.SizeOf(typeof(VertexType));
            bd.ByteWidth = obj.VertexStride * vertices.Length;
            bd.BindFlags = BindingFlag.VertexBuffer;
            bd.CPUAccessFlags = 0;
            bd.MiscFlags = 0;
            SubResourceData initData = new SubResourceData();
            initData.SystemMemory = ManagedDirectX.MathUtility.ArrayToByte<VertexType>(vertices);

            obj.VertexBuffer = device.DevicePointer.CreateBuffer(bd, initData);


            bd = new BufferDescription();

            bd.Usage = ManagedDirectX.Direct3D11.Resource11.Usage.Default;
            bd.ByteWidth = 2 * indices.Length;
            bd.BindFlags = BindingFlag.IndexBuffer;
            bd.CPUAccessFlags = 0;
            bd.MiscFlags = 0;
            initData = new SubResourceData();
            initData.SystemMemory = ManagedDirectX.MathUtility.ArrayToByte<short>(indices);

            obj.IndexBuffer = device.DevicePointer.CreateBuffer(bd, initData);

            obj.Layout = device.DevicePointer.CreateInputLayout(desc, effect.VertexShaderBuffer);

            obj.FaceCount = indices.Length / 3;

            obj.Materials = new List<Material3D>();

            return obj;
        }

        /// <summary>
        /// Set buffer in device
        /// </summary>
        /// <param name="manager">Device</param>
        public void SetBuffer(DeviceManager manager)
        {
            manager.DeviceContextPointer.IASetInputLayout(this.Layout);
            manager.DeviceContextPointer.IASetVertexBuffers(VertexBuffer, VertexStride);
            manager.DeviceContextPointer.IASetIndexBuffer(IndexBuffer, ManagedDirectX.DXGI.Format.R16_UINT);
            manager.DeviceContextPointer.IASetPrimitiveTopology(PrimitiveTopology.TriangleList);
        }

        /// <summary>
        /// Material
        /// </summary>
        /// <param name="materials">Materials</param>
        public void SetMaterial(Material3D[] materials)
        {
            int count = 0;
            for (int i = 0; i < materials.Length; i++)
            {
                materials[i].Attribute = materials[i].Attribute + count;
                count = materials[i].Attribute;
            }
            Materials = materials.ToList();
        }

        /// <summary>
        /// Draw
        /// </summary>
        /// <param name="manager">Manager</param>
        /// <param name="indicesCount">Indices count</param>
        public void Draw(DeviceManager manager, int indicesCount)
        {
            manager.DeviceContextPointer.DrawIndexed(indicesCount);
        }

        /// <summary>
        /// Draw
        /// </summary>
        /// <param name="manager">Device Manager</param>
        /// <param name="start">Start indices</param>
        /// <param name="indicesCount">Indices count</param>
        public void Draw(DeviceManager manager, int start, int indicesCount)
        {
            manager.DeviceContextPointer.DrawIndexed(indicesCount, start, 0);
        }

        /// <summary>
        /// Draw
        /// </summary>
        /// <param name="manager">Device manager</param>
        public void Draw(DeviceManager manager)
        {
            manager.DeviceContextPointer.PSSetShaderResources(new ShaderResourceView[] { Materials[0].Texture });
            Draw(manager, 0, Materials[0].Attribute);

            for (int i = 1; i < Materials.Count; i++)
            {
                manager.DeviceContextPointer.PSSetShaderResources(new ShaderResourceView[] { Materials[i].Texture });
                Draw(manager, Materials[i - 1].Attribute, Materials[i].Attribute - Materials[i - 1].Attribute);
            }
        }



        public static Object3D FromWavefrontOBJ(DeviceManager manager, EffectShader effect, string path)
        {
            WaveFrontModel[] models = WaveFrontModel.CreateFromObj(path);
            string directory = System.IO.Path.GetDirectoryName(path);

            Material3D[] mate = (from w in models
                                 select
                                     new Material3D(
                                     w.FaceCounts.First(),
                                     directory + "\\" + System.IO.Path.GetFileName(w.MeshMaterial.First().TextureName),
                                     manager)
                 ).ToArray();

            WaveFrontModel objM = models.First();
            for (int i = 1; i < models.Length; i++)
            {
                objM = objM + models[i];
            }

            InputElementDescription[] layout = new InputElementDescription[] 
            {
            new InputElementDescription("POSITION",0,Format.R32G32B32_FLOAT,0,0,InputClassification.VertexData,0),
            new InputElementDescription("NORMAL",0,Format.R32G32B32_FLOAT,0,12,InputClassification.VertexData,0),
            new InputElementDescription("TEXCOORD",0,Format.R32G32_FLOAT,0,24,InputClassification.VertexData,0)
            };

            Object3D model = Object3D.Create<Vertex>(objM.VertexData.ToArray(), objM.IndexData.ToArray(), layout, effect, manager);
            model.SetMaterial(mate);
            return model;
        }

    }
}
