﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.IO;
using ManagedDirectX;
using System.Globalization;

namespace D3D11Helper
{
    /// <summary>
    /// Load an obj model
    /// </summary>
    public class WaveFrontModel
    {
        private static CultureInfo infos = CultureInfo.InvariantCulture;

        /// <summary>
        /// Name of the model
        /// </summary>
        public string Name { get; private set; }

        /// <summary>
        /// Semantic
        /// </summary>
        public List<Semantic> Semantics { get; private set; }

        /// <summary>
        /// Vertex data
        /// </summary>
        public List<Vertex> VertexData { get; private set; }

        /// <summary>
        /// Index data
        /// </summary>
        public List<short> IndexData { get; private set; }

        /// <summary>
        /// Mesh material
        /// </summary>
        public List<Material> MeshMaterial { get; private set; }

        /// <summary>
        /// Face counts
        /// </summary>
        public List<int> FaceCounts { get; private set; }

        /// <summary>
        /// Semantic description
        /// </summary>
        public class Semantic
        {
            /// <summary>
            /// Name
            /// </summary>
            public String Name { get; set; }
            /// <summary>
            /// Element counts
            /// </summary>
            public int ElementCount { get; set; }
        }

        /// <summary>
        /// Material description
        /// </summary>
        public class Material
        {
            /// <summary>
            /// Name
            /// </summary>
            public string MaterialName;
            /// <summary>
            /// Ambient color
            /// </summary>
            public Vector3 Ambient;
            /// <summary>
            /// Diffuse color
            /// </summary>
            public Vector3 Diffuse;
            /// <summary>
            /// Specular color
            /// </summary>
            public Vector3 Specular;
            /// <summary>
            /// Shininess
            /// </summary>
            public float Shininess;
            /// <summary>
            /// Texture name
            /// </summary>
            public string TextureName;
        }

        /// <summary>
        /// Fuse 2 model
        /// </summary>
        /// <param name="a">A</param>
        /// <param name="b">B</param>
        /// <returns>Sum</returns>
        public static WaveFrontModel operator +(WaveFrontModel a, WaveFrontModel b)
        {
            WaveFrontModel c = new WaveFrontModel();

            c.Name = a.Name;
            c.VertexData.AddRange(a.VertexData);
            c.VertexData.AddRange(b.VertexData);
            c.Semantics.AddRange(a.Semantics);
            c.IndexData.AddRange(a.IndexData);

            c.IndexData.AddRange(
                (from i in b.IndexData select (short)(i + a.VertexData.Count)).ToArray());
            c.MeshMaterial.AddRange(a.MeshMaterial);
            c.MeshMaterial.AddRange(b.MeshMaterial);

            int tot = 0;
            foreach (int i in a.FaceCounts)
            {
                c.FaceCounts.Add(i + tot);
                tot += i;
            }

            foreach (int i in b.FaceCounts)
            {
                c.FaceCounts.Add(i + tot);
                tot += i;
            }

            return c;
        }

        private WaveFrontModel()
        {
            this.IndexData = new List<short>();
            this.Name = "";
            this.Semantics = new List<Semantic>();
            this.VertexData = new List<Vertex>();
            this.MeshMaterial = new List<Material>();
            this.FaceCounts = new List<int>();
        }

        /// <summary>
        /// Create obj from file
        /// </summary>
        /// <param name="filename">Filename</param>
        /// <returns>Models</returns>
        public static WaveFrontModel[] CreateFromObj(string filename)
        {
            List<WaveFrontModel> geom = new List<WaveFrontModel>();

            List<Vector3> positions = new List<Vector3>();
            List<Vector3> normals = new List<Vector3>();
            List<Vector2> textures = new List<Vector2>();
            List<int> faces = new List<int>();
            List<Material> materials = new List<Material>();
            List<string> lines = new List<string>();
            using (StreamReader reader = new StreamReader(filename))
            {
                for (; ; )
                {
                    string l = reader.ReadLine();
                    if (reader.EndOfStream)
                        break;

                    if (l.Contains("#") || string.IsNullOrEmpty(l.Trim()))
                        continue;

                    lines.Add(l);
                }
            }

            foreach (string line in lines)
            {
                if (line.Contains("v "))
                {
                    positions.Add(GetPosition(line));
                }
                else if (line.Contains("vn "))
                {
                    normals.Add(GetNormal(line));
                }
                else if (line.Contains("vt "))
                {
                    textures.Add(GetTexture(line));
                }
                else if (line.Contains("mtllib "))
                {
                    string path = System.IO.Path.GetDirectoryName(filename) + "\\" + line.Replace("mtllib", "").Trim();
                    materials.AddRange(LoadMaterial(path));
                }
            }

            string currentMesh = "default";
            Material currentMate = new Material();

            foreach (string line in lines)
            {
                if (line.Contains("f "))
                {
                    faces.AddRange(GetFace(line));
                }
                else if (line.Contains("usemtl "))
                {
                    if (faces.Count > 0)
                        geom.Add(CreateGeom(currentMesh, positions, normals, textures, faces, currentMate));

                    currentMate = (from m in materials where m.MaterialName == line.Replace("usemtl", "").Trim() select m).First();

                    faces.Clear();
                }
                else if (line.Contains("g "))
                {
                    if (faces.Count > 0)
                        geom.Add(CreateGeom(currentMesh, positions, normals, textures, faces, currentMate));

                    currentMesh = line.Replace("g", "").Trim();

                    faces.Clear();
                }
                else if (line.Contains("# "))
                {
                    //commento
                }
            }


            if (faces.Count > 0)
                geom.Add(CreateGeom(currentMesh, positions, normals, textures, faces, currentMate));

            return geom.ToArray();
        }

        private static WaveFrontModel CreateGeom(string name, List<Vector3> position, List<Vector3> normals, List<Vector2> texture, List<int> faces, Material mate)
        {
            WaveFrontModel geom = new WaveFrontModel();
            geom.Name = name;
            geom.Semantics = new List<Semantic>();
            geom.VertexData = new List<Vertex>();
            geom.IndexData = new List<short>();

            int stride = 0;


            if (position.Count > 0)
            {
                geom.Semantics.Add(new Semantic()
                {
                    Name = "POSITION",
                    ElementCount = 3
                });
                stride++;
            }

            if (normals.Count > 0)
            {
                geom.Semantics.Add(new Semantic()
                {
                    Name = "NORMAL",
                    ElementCount = 3
                });
                stride++;
            }

            if (texture.Count > 0)
            {
                geom.Semantics.Add(new Semantic()
                {
                    Name = "TEXCOORD",
                    ElementCount = 2
                });
                stride++;
            }

            int vertexCount = faces.Count / stride;

            int k = 0;


            for (int i = 0; i < vertexCount; i++)
            {
                Vertex v = new Vertex();
                if (position.Count > 0)
                {
                    v.Position = position[faces[k] - 1];
                    k++;
                }

                if (texture.Count > 0)
                {
                    v.Texture = texture[faces[k] - 1];
                    k++;
                }

                if (normals.Count > 0)
                {
                    v.Normal = normals[faces[k] - 1];
                    k++;
                }
                geom.VertexData.Add(v);

            }

            //triangolarizzazione
            List<Vertex> temp = new List<Vertex>();
            List<short> iTemp = new List<short>();

            for (int i = 0; i < geom.VertexData.Count; i++)
            {
                int n = temp.IndexOf(geom.VertexData[i]);
                if (n < 0)
                {
                    temp.Add(geom.VertexData[i]);
                    iTemp.Add((short)(temp.Count - 1));
                }
                else
                {
                    iTemp.Add((short)n);
                }
            }
            geom.VertexData = temp;
            geom.IndexData = iTemp;

            //for (int i = 0; i < (geom.VertexData.Count); i++)
            //{
            //    geom.IndexData.Add((short)i);
            //}

            geom.MeshMaterial.Add(mate);
            geom.FaceCounts.Add(geom.IndexData.Count);
            return geom;

        }

        private static string[] Parts(string name, string line)
        {
            return line.Replace(name, "").Trim().Split(' ');
        }

        private static Vector3 GetPosition(string line)
        {
            string[] parts = Parts("v", line);
            return new Vector3(
                float.Parse(parts[0], infos),
                float.Parse(parts[1], infos),
                float.Parse(parts[2], infos));
        }

        private static Vector3 GetNormal(string line)
        {
            string[] parts = Parts("vn", line);
            return new Vector3(
                float.Parse(parts[0], infos),
                float.Parse(parts[1], infos),
                float.Parse(parts[2], infos));
        }

        private static Vector2 GetTexture(string line)
        {
            string[] parts = Parts("vt", line);
            return new Vector2(
                float.Parse(parts[0], infos),
                1.0F - float.Parse(parts[1], infos));
        }

        private static int[] GetFace(string line)
        {
            string[] parts = Parts("f", line);
            List<int> lists = new List<int>();
            foreach (String s in parts)
            {
                lists.AddRange((from ss in s.Split('/') select int.Parse(ss)).ToArray());
            }
            return lists.ToArray();
        }

        private static WaveFrontModel Create(StreamReader reader)
        {
            WaveFrontModel geom = new WaveFrontModel();

            return geom;
        }


        private static Material[] LoadMaterial(String filename)
        {
            Material current = null;
            List<Material> materials = new List<Material>();
            using (StreamReader reader = new StreamReader(filename))
            {


                for (; ; )
                {
                    if (reader.EndOfStream)
                        break;
                    string line = reader.ReadLine();



                    if (line.Contains("newmtl"))
                    {
                        if (current != null)
                            materials.Add(current);

                        current = new Material();
                        current.MaterialName = line.Replace("newmtl", "").Trim();
                    }
                    else if (line.Contains("map_Kd"))
                    {
                        current.TextureName = line.Replace("map_Kd", "").Trim();
                    }
                    else if (line.Contains("map_Ka"))
                    {

                    }
                    else if (line.Contains("Ka"))
                    {
                        float[] val = (from s in line.Replace("Ka", "").Trim().Split(' ') select float.Parse(s, infos)).ToArray();
                        current.Ambient = new Vector3(val[0], val[1], val[2]);
                    }
                    else if (line.Contains("Kd"))
                    {
                        float[] val = (from s in line.Replace("Kd", "").Trim().Split(' ') select float.Parse(s, infos)).ToArray();
                        current.Diffuse = new Vector3(val[0], val[1], val[2]);
                    }
                    else if (line.Contains("Ks"))
                    {
                        float[] val = (from s in line.Replace("Ks", "").Trim().Split(' ') select float.Parse(s, infos)).ToArray();
                        current.Specular = new Vector3(val[0], val[1], val[2]);
                    }
                    else if (line.Contains("Ns"))
                    {
                        float[] val = (from s in line.Replace("Ns", "").Trim().Split(' ') select float.Parse(s, infos)).ToArray();
                        current.Shininess = val.First();
                    }


                }
            }

            if (current != null)
                materials.Add(current);

            return materials.ToArray();
        }
    }

}



